package com.fullbattery.osgi.bootloader;

import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.List;
import java.util.TooManyListenersException;

import javax.print.attribute.standard.Chromaticity;

import com.fullbattery.osgi.bootloader.intelhex.HexParser;
import com.fullbattery.osgi.bootloader.intelhex.HexParserException;

public class Bootloader implements SerialPortEventListener {
	private static final int BL_COMMAND = 0xA5; // Command Sequence start
	private static final int BL_ESCAPE = BL_COMMAND;

	private static final int BL_CONNECT = 0xA6; // connection established
	private static final int BL_BADCOMMAND = 0xA7; // command not supported
	private static final int BL_ANSWER = 0xA8; // followed by length byte
	private static final int BL_CONTINUE = 0xA9;
	private static final int BL_SUCCESS = 0xAA;
	private static final int BL_FAIL = 0xAB;
	private static final int BL_ESC_SHIFT = 0x80; // offset escape char
	private static final int BL_FF = 0xFF; // offset escape char
	private static final int BL_PROGEND = BL_ESC_SHIFT;

	private static final int BL_REVISION = 0; // get Bootloader revision
	private static final int BL_BUFFSIZE = 1; // get buffer size
	private static final int BL_SIGNATURE = 2; // get target signature
	private static final int BL_USERFLASH = 3; // get user flash size
	private static final int BL_PROGRAMM = 4; // programm flash
	private static final int BL_START = 5; // start application
	private static final int BL_CHECK_CRC = 6; // CRC OK
	private static final int BL_VERIFY = 7; // verify

	private static final String CMD_RESET = "reset()Q";
	private static final String CMD_PASSWD = "Peda";
	private static final int MAX_ANSWER_LENGTH = 10;

	/**
	 * Maximale Wartezeit in Miiliskunden für eine Antwort des Bootloader auf
	 * ein Befehl
	 */
	private static final long TIME_OUT = 700;

	/**
	 * Maximale Wartezeit in Miiliskunden für den Versuch den Bootloader zu
	 * kontaktieren
	 */
	private static final long TIME_OUT_BOOTLOADER_CONNECT = 1000;

	/**
	 * Der Serielle Port mit dem wir sprechen
	 */
	protected SerialPort serialPort;

	/**
	 * Der Inputstream über den wir Daten empfangen
	 */
	protected InputStream in;

	/**
	 * Der Aoutputstream auf dem wir schreiben
	 */
	protected OutputStream out;

	/**
	 * Enthält die zuletzt empfangene Antwort
	 */
	protected int[] answer = new int[MAX_ANSWER_LENGTH];

	/**
	 * Speichert die empfangenen Daten zwischen
	 */
	protected byte[] buffer = new byte[MAX_ANSWER_LENGTH];

	/**
	 * Connectiert sich mit der Seriellen Schnittstelle und hört auf Events
	 * 
	 * @param whichPort
	 * @param whichSpeed
	 * @throws NoSuchPortException
	 * @throws PortInUseException
	 * @throws UnsupportedCommOperationException
	 * @throws IOException
	 * @throws TooManyListenersException
	 */
	public void connectSerialPort(String whichPort, int whichSpeed) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException,
			TooManyListenersException {
		// find the port
		CommPortIdentifier portId = CommPortIdentifier.getPortIdentifier(whichPort);

		// open the port
		serialPort = (SerialPort) portId.open("gu_java_serial" + whichPort, 1000);

		// configure the port
		serialPort.setSerialPortParams(whichSpeed, SerialPort.DATABITS_8, SerialPort.STOPBITS_2, SerialPort.PARITY_NONE);
		in = serialPort.getInputStream();
		out = serialPort.getOutputStream();

		// we could read from "in" in a separate thread, but the API gives us
		// events
		serialPort.addEventListener(this);
		serialPort.notifyOnDataAvailable(true);
	}

	/**
	 * Läd die neue Software auf die Ladekarte
	 * 
	 * @param newSoftware
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws IOException
	 * @throws ProgrammToLargeException
	 * @throws NoAnswerException
	 * @throws InterruptedException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 * @throws ProgrammToLargeException
	 * @throws NoAnswerException
	 * @throws ProgrammToLargeException
	 * @throws HexParserException
	 */
	public void uploadSoftware(String newSoftware) throws IOException, InterruptedException, NoAnswerException, ProgrammToLargeException, HexParserException {

		System.out.println("main vor connect CRC  = " + crc);

		if (connect()) {
			System.out.println("main nach connect CRC  = " + crc);

			initialCommand();

			System.out.println("main nach initial command CRC  = " + crc);

			answer = new int[MAX_ANSWER_LENGTH];

			CRC_STATUS crc_on;
			crc_on = check_crc(); // 2 = no CRC support

			System.out.println("main nach ersten check_crc CRC  = " + crc);

			//crc_on = CRC_STATUS.NO_CRC;

			readInfo();
			System.out.println("main nach readInfo CRC  = " + crc);

			if ((crc_on != CRC_STATUS.NO_CRC)) {
				crc_on = check_crc();
				if (crc_on == CRC_STATUS.CRC_FAIL) {
					throw new RuntimeException("CRC Error");
				}
			}
			
			
			// Die HexDaten aus dem HexFile laden
			HexParser parser = new HexParser();
			parser.loadHexFile(newSoftware);
			List<Integer> hexData = parser.getHexData();

			if (hexData.size() > userFlash) {
				throw new ProgrammToLargeException();
			}

			/* Programmieren der Software */
			programm(hexData, false);

			/* Verify der Software */
			//programm(hexData, true);

			/* CRC Prüfung */
			if (crc_on == CRC_STATUS.CRC_OK) {
				check_crc();
			}

			/* Start der Software */
			System.out.println("Start the Program");
			endProgramming();
		}
	}

	/**
	 * Beendet die Programmierung und startet das Programm
	 * 
	 * @throws IOException
	 */
	public void endProgramming() throws IOException {
		sendcommand(BL_PROGEND);
		sendcommand(BL_COMMAND);
		sendcommand(BL_START);
	}

	/**
	 * Versucht den Bootloader zu connectieren
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 */
	protected boolean connect() throws IOException, InterruptedException {
		isConnected = false;

		// Zuerste senden wir ein Reset Command
		out.write(CMD_RESET.getBytes());

		// Diese Zeit versuchen wir einen Connect zu bekommen
		long maxMilli = System.currentTimeMillis() + TIME_OUT_BOOTLOADER_CONNECT;

		// Solange noch keine Connection steht und die Zet noch nicht abgelaufen
		// ist
		// versuchen wir es weiter
		while (!isConnected && (System.currentTimeMillis() < maxMilli)) {
			for (byte b : CMD_PASSWD.getBytes()) {
				send(b);
			}
			// out.write(CMD_PASSWD.getBytes());

			Thread.sleep(50);
			// System.out.println("Answer="+answer.toString());
			// System.out.println("Compare '"+answer[0]+"'<->'"+BL_CONNECT+"'");
			if (answer[0] == BL_CONNECT) {
				isConnected = true;
				System.out.println("CONNECT");
				return true;
			}else{
				send(BL_FF);
			}
			answer = new int[MAX_ANSWER_LENGTH];
		}

		return false;
	}


	/**
	 * Versucht den Bootloader zu kontaktieren
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private void initialCommand() throws IOException, InterruptedException, NoAnswerException {
		isAnswerArrived = false;
		sendcommand(BL_COMMAND);
		waitForAnswer();

		if (answer[0] != BL_SUCCESS) {
			throw new RuntimeException("Initial Command Fehlgeschlagen");
		}
	}

	/**
	 * Ermittelt die Version des Bootloaders
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private void getBootloaderVersion() throws IOException, InterruptedException, NoAnswerException {
		sendcommand(BL_REVISION);
		waitForAnswer();
		revision = readVal();
	}

	/**
	 * Ermittelt die Puffergröße
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private void getUserFlashSize() throws IOException, InterruptedException, NoAnswerException {
		sendcommand(BL_USERFLASH);
		waitForAnswer();
		userFlash = readVal();
	}

	private int userFlash;

	private int revision;

	private int bufferSize;

	/**
	 * Wartet auf eine Antwort des Bootloaders
	 * 
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	protected void waitForAnswer() throws InterruptedException, NoAnswerException {
		long maxMilli = System.currentTimeMillis() + TIME_OUT;
		while (!isAnswerArrived && (System.currentTimeMillis() < maxMilli)) {
			Thread.sleep(50);
		}

		if (isAnswerArrived) {
			for (int i = 0; i < answer.length; i++) {
				System.out.print(answer[i] + " ");
			}
			System.out.println("");
			for (int i = 0; i < answer.length; i++) {
				System.out.print(Integer.toHexString(answer[i]) + " ");
			}
			System.out.println("");
		} else {
			throw new NoAnswerException();
		}
	}

	protected boolean isAnswerArrived = false;

	/**
	 * Schließt die Seriell Schnittstelle
	 */
	public void disconnect() {
		if (serialPort != null) {
			serialPort.removeEventListener();
			serialPort.close();
		}
	}

	/**
	 * Wird auf true gesetzt sobald der Bootloader erreicht werden konnte
	 */
	protected boolean isConnected = false;

	/**
	 * Wird aufgerufen wenn Daten an der Schnittstelle anliegen
	 */
	public void serialEvent(SerialPortEvent event) {
		if (event.getEventType() == SerialPortEvent.DATA_AVAILABLE) {
			try {
				// Ermitteln wieviele Daten anliegen
				int length = in.available();

				length = length < MAX_ANSWER_LENGTH ? length : MAX_ANSWER_LENGTH;
				// System.out.println("lenth=" + length);
				int readBytes = in.read(buffer, 0, length);
				// System.out.println("read=" + readBytes);

				for (int i = 0; i < readBytes; i++) {
					int val = (0xFF & buffer[i]);
					// answer.append(Integer.toHexString(val));
					// answer.append(val);
					// String.valueOf((char)val);
					answer[i] = val;
				}
				isAnswerArrived = true;
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}
	}

	/* -------------------------------------------- */
	/* -------------------------------------------- */
	/* -------------------------------------------- */
	/* -------------------------------------------- */
	/* -------------------------------------------- */

	/**
	 * Speichert den aktuellen CRC Wert
	 */
	private int crc;

	/**
	 * fügt dem aktuellen CRC ein neues Zeichen hinzu
	 * 
	 * @param d
	 */
	private void calcCrc(int d) {
		int tmpCrc = crc ^ d;

		for (int i = 0; i < 8; i++) {
			if ((tmpCrc & 0x1) != 0) {
				tmpCrc = (tmpCrc >> 1) ^ 0xA001;
			} else {
				tmpCrc = tmpCrc >> 1;
			}

		}
		crc = tmpCrc & 0xffff;
	}

	/**
	 * Sendet das Byte und Calculiert die Checksumme
	 * 
	 * @param c
	 * @throws IOException
	 */
	private void send(int c) throws IOException {
		isAnswerArrived = false;
		answer = new int[10];
		out.write((char) c);
		System.out.println("SENDEN Char = " + c);
		calcCrc(c); // calculate transmit CRC
		System.out.println("SENDEN CRC  = " + crc);
	}

	/**
	 * Sendet ein Command
	 * 
	 * @param cmd
	 *            Das zu sendende Command
	 * @throws IOException
	 */
	private void sendcommand(int cmd) throws IOException {
		send(BL_COMMAND);
		send(cmd);
	}

	/**
	 * Gibt den CRC_STATUS zurück
	 * 
	 * @param flag
	 * @return
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private CRC_STATUS check_crc() throws IOException, InterruptedException, NoAnswerException {
		int i;

		sendcommand(BL_CHECK_CRC);
		Thread.sleep(100);

		if (isAnswerArrived) {
			if (answer[0] == BL_BADCOMMAND) {
				return CRC_STATUS.NO_CRC;
			}
		}

		int tmpCrc = crc;
		System.out.println(Integer.toHexString(tmpCrc));

		send(tmpCrc & 0xff);
		send(tmpCrc >> 8);

		waitForAnswer();

		i = answer[0];
		switch (i) {
		case BL_SUCCESS:
			return CRC_STATUS.CRC_OK;

		default:
			return CRC_STATUS.CRC_FAIL;
		}
	}

	/**
	 * Wartet bis ein Anwort da ist und gibt das erste Byte zurück
	 * 
	 * @return
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private int empfang() throws InterruptedException, NoAnswerException {
		waitForAnswer();
		return answer[0];
	}

	/**
	 * Liest die Antwort aus die mehr als ein Byte hat. Byte 2 enthält die
	 * Anzahl der noch folgenden Bytes. Das Letzte Byte muss immer BL_SUCCESS
	 * sein.
	 * 
	 * @return
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private int readVal() throws IOException, InterruptedException, NoAnswerException {
		int result = 0;
		int byteCount;
		int offset = 2;

		if (answer[0] == BL_ANSWER) {

			// Die Anzahl der Bytes die zu einem long werden
			byteCount = answer[1] - 1;

			for (int i = 0; i < byteCount; i++) {
				int shift = (byteCount - (i + 1)) * 8;
				result = result | answer[i + offset] << shift;
			}

			if (answer[byteCount + offset] == BL_SUCCESS) {
				return result;
			}
		}
		throw new RuntimeException("getBufferSize geht nicht");
	}

	private ControllerType target;

	/**
	 * Ermittelt die Puffergröße
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private void getBufferSize() throws IOException, InterruptedException, NoAnswerException {
		isAnswerArrived = false;
		sendcommand(BL_BUFFSIZE);
		waitForAnswer();

		bufferSize = readVal();
	}

	/**
	 * Ermittelt die Puffergröße
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private void getTarget() throws IOException, InterruptedException, NoAnswerException {
		isAnswerArrived = false;
		sendcommand(BL_SIGNATURE);
		waitForAnswer();

		int i = readVal();

		if (i == 0x1e9007) {
			target = ControllerType.AT_TINY_13;
		} else if (i == 0x1e910A) {
			target = ControllerType.AT_TINY_2313;
		} else if (i == 0x1e9206) {
			target = ControllerType.AT_TINY_45;
		} else if (i == 0x1e9205) {
			target = ControllerType.AT_MEGA_48;
		} else if (i == 0x1e9307) {
			target = ControllerType.AT_MEGA_8;
		} else if (i == 0x1e9403) {
			target = ControllerType.AT_MEGA_16;
		} else if (i == 0x1e9406) {
			target = ControllerType.AT_MEGA_168;
		} else if (i == 0x1e9502) {
			target = ControllerType.AT_MEGA_32;
		} else if (i == 0x1e9609) {
			target = ControllerType.AT_MEGA_644;
		} else if (i == 0x1e9802) {
			target = ControllerType.AT_MEGA_2561;
		} else {
			target = ControllerType.UNKNOWN;
		}

	}

	/**
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private void readInfo() throws IOException, InterruptedException, NoAnswerException {
		System.out.println("Get Bootloader Version");
		getBootloaderVersion();

		System.out.println("Get Target");
		getTarget();

		System.out.println("Get Buff Size");
		getBufferSize();

		System.out.println("Get User Flash Size");
		getUserFlashSize();
	}

	private void programm(List<Integer> hexData, boolean isVerify) throws IOException, InterruptedException, NoAnswerException {

		if (!isVerify) {
			System.out.println("Start Programming");
			sendcommand(BL_PROGRAMM);
		} else {
			System.out.println("Start Verifing");
			sendcommand(BL_VERIFY);
			Thread.sleep(300);
			if (isAnswerArrived) {
				if (answer[0] == BL_BADCOMMAND) {
					throw new RuntimeException("Verify not available");
				}
			}
		}

		int bytesWritten = 0;
		int bufferCounter = bufferSize;

		for (int data : hexData) {
			System.out.println("Write byte '" + Integer.toHexString(bytesWritten) + "' = " + Integer.toHexString(data));
			if ((data == BL_ESCAPE) || (data == 0x13)) {
				send(BL_ESCAPE);
				send(data + BL_ESC_SHIFT);
				//send(BL_ESC_SHIFT);
			} else {
				send(data);
			}

			if (--bufferCounter == 0) {
				System.out.println("Buffer Groesse erreicht");
				if (!isVerify && empfang() != BL_CONTINUE) {
					throw new RuntimeException("Programming Failed");
				}
				bufferCounter = bufferSize;
			}

			bytesWritten++;
		}

		// so, jetzt sind alle bytes gesendet.
		send(BL_ESCAPE);
		send(BL_ESC_SHIFT);
		if (empfang() == BL_SUCCESS) {
			System.out.println(" successful");
		} else {
			throw new RuntimeException("Programming Failed");
		}

	}

	/**
	 * Führt die eigentliche Programmierung und Verifikation durch
	 * 
	 * @param dataInputStream
	 * @param isVerify
	 * @throws IOException
	 * @throws InterruptedException
	 * @throws NoAnswerException
	 */
	private void old_programm(InputStream dataInputStream, boolean isVerify) throws IOException, InterruptedException, NoAnswerException {

		if (!isVerify) {
			sendcommand(BL_PROGRAMM);
		} else {
			sendcommand(BL_VERIFY);
			Thread.sleep(300);
			if (isAnswerArrived) {
				if (answer[0] == BL_BADCOMMAND) {
					throw new RuntimeException("Verify not available");
				}
			}
		}

		int bytesWritten = 0;
		int data;
		data = dataInputStream.read();
		int bufferCounter = bufferSize;
		while (data != -1) {
			System.out.println("Write byte '" + Integer.toHexString(bytesWritten) + "' = " + Integer.toHexString(data));
			if ((data == BL_ESCAPE) || (data == 0x13)) {
				send(BL_ESCAPE);
				send(data);
				send(BL_ESC_SHIFT);
			} else {
				send(data);
			}

			if (--bufferCounter == 0) {
				System.out.println("Buffer Groesse erreicht");
				if (!isVerify && empfang() != BL_CONTINUE) {
					throw new RuntimeException("Programming Failed");
				}
				bufferCounter = bufferSize;
			}

			data = dataInputStream.read();
			bytesWritten++;
		}

		// so, jetzt sind alle bytes gesendet.
		send(BL_ESCAPE);
		send(BL_ESC_SHIFT);
		if (empfang() == BL_SUCCESS) {
			System.out.println(" successful");
		} else {
			throw new RuntimeException("Programming Failed");
		}
	}

}
